@@ -1,3 +1,70 @@
+{-| For further information see the <http://www.cse.chalmers.se/~rjmh/QuickCheck/manual.html QuickCheck manual>.
+
+To use QuickCheck to check a property, first define a function
+expressing that property (functions expressing properties under test
+tend to be prefixed with @prop_@). Testing that @n + m = m + n@ holds
+for @Integer@s one might write:
+
+@
+import Test.QuickCheck
+
+prop_commutativeAdd :: Integer -> Integer -> Bool
+prop_commutativeAdd n m = n + m == m + n
+@
+
+and testing:
+
+>>> quickcheck prop_commutativeAdd
++++ OK, passed 100 tests.
+
+which tests @prop_commutativeAdd@ on 100 random @(Integer, Integer)@ pairs.
+
+'verboseCheck' can be used to see the actual values generated:
+
+>>> verboseCheck prop_commutativeAdd
+Passed:
+0
+0
+  …98 tests omitted…
+Passed:
+-68
+6
++++ OK, passed 100 tests.
+
+and if more than 100 tests are needed the number of tests can be
+increased by updating the 'stdArgs' record:
+
+>>> quickCheckWith stdArgs { maxSuccess = 500 } prop_commutativeAdd
++++ OK, passed 500 tests.
+
+To let QuickCheck generate values of your own data type an 'Arbitrary'
+instance must be defined:
+
+@
+data Point = MkPoint Int Int deriving Eq
+
+instance Arbitrary Point where
+  arbitrary = do
+    x <- 'arbitrary'
+    y <- arbitrary
+    return (MkPoint x y)
+
+swapPoint :: Point -> Point
+swapPoint (MkPoint x y) = MkPoint y x
+
+-- swapPoint . swapPoint = id
+prop_swapInvolution point = swapPoint (swapPoint point) == point
+@
+
+>>> quickCheck prop_swapInvolution
++++ OK, passed 100 tests.
+
+See "Test.QuickCheck.Function" for generating random shrinkable,
+showable functions used for testing higher-order functions and
+"Test.QuickCheck.Monadic" for testing impure or monadic code
+(e.g. effectful code in 'IO').
+
+-}
 {-# LANGUAGE CPP #-}
 #ifndef NO_SAFE_HASKELL
 {-# LANGUAGE Safe #-}
@@ -37,12 +104,15 @@ module Test.QuickCheck
   , growingElements
   , sized
   , resize
+  , scale
   , suchThat
   , suchThatMaybe
   , listOf
   , listOf1
   , vectorOf
   , infiniteListOf
+  , shuffle
+  , sublistOf
     -- ** Generators which use Arbitrary
   , vector
   , orderedList
@@ -59,6 +129,7 @@ module Test.QuickCheck
 
     -- ** Helper functions for implementing arbitrary
   , arbitrarySizedIntegral
+  , arbitrarySizedNatural
   , arbitrarySizedFractional
   , arbitrarySizedBoundedIntegral
   , arbitraryBoundedIntegral
@@ -66,6 +137,7 @@ module Test.QuickCheck
   , arbitraryBoundedEnum
     -- ** Helper functions for implementing shrink
 #ifndef NO_GENERICS
+  , genericCoarbitrary
   , genericShrink
   , subterms
   , recursivelyShrink
