@@ -54,6 +54,9 @@ tryEvaluate x = return (Right x)
 tryEvaluateIO :: IO a -> IO (Either AnException a)
 tryEvaluateIO m = fmap Right m
 
+evaluate :: a -> IO a
+evaluate x = x `seq` return x
+
 isInterrupt :: AnException -> Bool
 isInterrupt _ = False
 
@@ -80,7 +83,10 @@ tryEvaluateIO :: IO a -> IO (Either AnException a)
 tryEvaluateIO m = E.try (m >>= E.evaluate)
 --tryEvaluateIO m = Right `fmap` m
 
--- Test if an exception was a ^C.
+evaluate :: a -> IO a
+evaluate = E.evaluate
+
+-- | Test if an exception was a @^C@.
 -- QuickCheck won't try to shrink an interrupted test case.
 isInterrupt :: AnException -> Bool
 
