@@ -19,10 +19,17 @@ import System.Random
 import Control.Monad
   ( liftM
   , ap
+  , replicateM
+  , filterM
   )
 
 import Control.Applicative
   ( Applicative(..)
+  , (<$>)
+  )
+
+import Control.Arrow
+  ( second
   )
 
 import Test.QuickCheck.Random
@@ -69,13 +76,20 @@ sized f = MkGen (\r n -> let MkGen m = f n in m r n)
 -- | Overrides the size parameter. Returns a generator which uses
 -- the given size instead of the runtime-size parameter.
 resize :: Int -> Gen a -> Gen a
-resize n (MkGen m) = MkGen (\r _ -> m r n)
+resize n _ | n < 0 = error "Test.QuickCheck.resize: negative size"
+resize n (MkGen g) = MkGen (\r _ -> g r n)
+
+-- | Adjust the size parameter, by transforming it with the given
+-- function.
+scale :: (Int -> Int) -> Gen a -> Gen a
+scale f g = sized (\n -> resize (f n) g)
 
 -- | Generates a random element in the given inclusive range.
 choose :: Random a => (a,a) -> Gen a
 choose rng = MkGen (\r _ -> let (x,_) = randomR rng r in x)
 
--- | Run a generator.
+-- | Run a generator. The size passed to the generator is always 30;
+-- if you want another size then you should explicitly use 'resize'.
 generate :: Gen a -> IO a
 generate (MkGen g) =
   do r <- newQCGen
@@ -135,6 +149,20 @@ elements :: [a] -> Gen a
 elements [] = error "QuickCheck.elements used with empty list"
 elements xs = (xs !!) `fmap` choose (0, length xs - 1)
 
+-- | Generates a random subsequence of the given list.
+sublistOf :: [a] -> Gen [a]
+sublistOf xs = filterM (\_ -> choose (False, True)) xs
+
+-- | Generates a random permutation of the given list.
+shuffle :: [a] -> Gen [a]
+shuffle [] = return []
+shuffle xs = do
+  (y, ys) <- elements (selectOne xs)
+  (y:) <$> shuffle ys
+  where
+    selectOne [] = []
+    selectOne (y:ys) = (y,ys) : map (second (y:)) (selectOne ys)
+
 -- | Takes a list of elements of increasing size, and chooses
 -- among an initial segment of the list. The size of this initial
 -- segment increases with the size parameter.
@@ -171,7 +199,7 @@ listOf1 gen = sized $ \n ->
 
 -- | Generates a list of the given length.
 vectorOf :: Int -> Gen a -> Gen [a]
-vectorOf k gen = sequence [ gen | _ <- [1..k] ]
+vectorOf = replicateM
 
 -- | Generates an infinite list.
 infiniteListOf :: Gen a -> Gen [a]
