@@ -13,7 +13,7 @@ import Test.QuickCheck.Gen.Unsafe
 import Test.QuickCheck.Arbitrary
 import Test.QuickCheck.Text( showErr, isOneLine, putLine )
 import Test.QuickCheck.Exception
-import Test.QuickCheck.State
+import Test.QuickCheck.State hiding (labels)
 
 #ifndef NO_TIMEOUT
 import System.Timeout(timeout)
@@ -21,6 +21,10 @@ import System.Timeout(timeout)
 import Data.Maybe
 import Control.Applicative
 import Control.Monad
+import qualified Data.Map as Map
+import Data.Map(Map)
+import qualified Data.Set as Set
+import Data.Set(Set)
 
 --------------------------------------------------------------------------
 -- fixities
@@ -168,23 +172,23 @@ instance Monad Rose where
   -- k must be total
   m >>= k  = joinRose (fmap k m)
 
--- Execute the "IORose" bits of a rose tree, returning a tree
+-- | Execute the "IORose" bits of a rose tree, returning a tree
 -- constructed by MkRose.
 reduceRose :: Rose Result -> IO (Rose Result)
 reduceRose r@(MkRose _ _) = return r
 reduceRose (IORose m) = m >>= reduceRose
 
--- Apply a function to the outermost MkRose constructor of a rose tree.
+-- | Apply a function to the outermost MkRose constructor of a rose tree.
 -- The function must be total!
 onRose :: (a -> [Rose a] -> Rose a) -> Rose a -> Rose a
 onRose f (MkRose x rs) = f x rs
 onRose f (IORose m) = IORose (fmap (onRose f) m)
 
--- Wrap a rose tree in an exception handler.
+-- | Wrap a rose tree in an exception handler.
 protectRose :: IO (Rose Result) -> IO (Rose Result)
 protectRose = protect (return . exception "Exception")
 
--- Wrap all the Results in a rose tree in exception handlers.
+-- | Wrap all the Results in a rose tree in exception handlers.
 protectResults :: Rose Result -> Rose Result
 protectResults = onRose $ \x rs ->
   IORose $ do
@@ -208,22 +212,11 @@ data Result
   , reason       :: String            -- ^ a message indicating what went wrong
   , theException :: Maybe AnException -- ^ the exception thrown, if any
   , abort        :: Bool              -- ^ if True, the test should not be repeated
-  , stamp        :: [(String,Int)]    -- ^ the collected values for this test case
+  , labels       :: Map String Int    -- ^ all labels used by this property
+  , stamp        :: Set String        -- ^ the collected values for this test case
   , callbacks    :: [Callback]        -- ^ the callbacks for this test case
   }
 
-result :: Result
-result =
-  MkResult
-  { ok           = undefined
-  , expect       = True
-  , reason       = ""
-  , theException = Nothing
-  , abort        = False
-  , stamp        = []
-  , callbacks    = []
-  }
-
 exception :: String -> AnException -> Result
 exception msg err
   | isDiscard err = rejected
@@ -238,14 +231,23 @@ formatException msg err = msg ++ ":" ++ format (show err)
 protectResult :: IO Result -> IO Result
 protectResult = protect (exception "Exception")
 
-succeeded :: Result
-succeeded = result{ ok = Just True }
-
-failed :: Result
-failed = result{ ok = Just False }
-
-rejected :: Result
-rejected = result{ ok = Nothing }
+succeeded, failed, rejected :: Result
+(succeeded, failed, rejected) =
+  (result{ ok = Just True },
+   result{ ok = Just False },
+   result{ ok = Nothing })
+  where
+    result =
+      MkResult
+      { ok           = undefined
+      , expect       = True
+      , reason       = ""
+      , theException = Nothing
+      , abort        = False
+      , labels       = Map.empty
+      , stamp        = Set.empty
+      , callbacks    = []
+      }
 
 --------------------------------------------------------------------------
 -- ** Lifting and mapping functions
@@ -301,7 +303,7 @@ counterexample s =
     res <- tryEvaluateIO (putLine (terminal st) s)
     case res of
       Left err ->
-        putLine (terminal st) (formatException "Exception thrown by generator" err)
+        putLine (terminal st) (formatException "Exception thrown while printing test case" err)
       Right () ->
         return ()
 
@@ -364,17 +366,22 @@ classify :: Testable prop =>
          -> prop -> Property
 classify b s = cover b 0 s
 
--- | Checks that at least the given proportion of the test cases belong
--- to the given class.
+-- | Checks that at least the given proportion of /successful/ test
+-- cases belong to the given class. Discarded tests (i.e. ones
+-- with a false precondition) do not affect coverage.
 cover :: Testable prop =>
          Bool   -- ^ @True@ if the test case belongs to the class.
       -> Int    -- ^ The required percentage (0-100) of test cases.
       -> String -- ^ Label for the test case class.
       -> prop -> Property
-cover True n s = n `seq` s `listSeq` (mapTotalResult $ \res -> res { stamp = (s,n) : stamp res })
+cover x n s =
+  x `seq` n `seq` s `listSeq`
+  mapTotalResult $
+    \res -> res {
+      labels = Map.insertWith max s n (labels res),
+      stamp = if x then Set.insert s (stamp res) else stamp res }
   where [] `listSeq` z = z
         (x:xs) `listSeq` z = x `seq` xs `listSeq` z
-cover False _ _ = property
 
 -- | Implication for properties: The resulting property holds if
 -- the first argument is 'False' (in which case the test case is discarded),
@@ -440,20 +447,20 @@ conjoin :: Testable prop => [prop] -> Property
 conjoin ps =
   MkProperty $
   do roses <- mapM (fmap unProp . unProperty . property) ps
-     return (MkProp (conj [] roses))
+     return (MkProp (conj id roses))
  where
-  conj cbs [] =
-    MkRose succeeded{callbacks = cbs} []
+  conj k [] =
+    MkRose (k succeeded) []
 
-  conj cbs (p : ps) = IORose $ do
+  conj k (p : ps) = IORose $ do
     rose@(MkRose result _) <- reduceRose p
     case ok result of
       _ | not (expect result) ->
         return (return failed { reason = "expectFailure may not occur inside a conjunction" })
-      Just True -> return (conj (cbs ++ callbacks result) ps)
+      Just True -> return (conj (addLabels result . addCallbacks result . k) ps)
       Just False -> return rose
       Nothing -> do
-        rose2@(MkRose result2 _) <- reduceRose (conj (cbs ++ callbacks result) ps)
+        rose2@(MkRose result2 _) <- reduceRose (conj (addCallbacks result . k) ps)
         return $
           -- Nasty work to make sure we use the right callbacks
           case ok result2 of
@@ -461,6 +468,12 @@ conjoin ps =
             Just False -> rose2
             Nothing -> rose2
 
+  addCallbacks result r =
+    r { callbacks = callbacks result ++ callbacks r }
+  addLabels result r =
+    r { labels = Map.unionWith max (labels result) (labels r),
+        stamp = Set.union (stamp result) (stamp r) }
+
 -- | Disjunction: 'p1' '.||.' 'p2' passes unless 'p1' and 'p2' simultaneously fail.
 (.||.) :: (Testable prop1, Testable prop2) => prop1 -> prop2 -> Property
 p1 .||. p2 = disjoin [property p1, property p2]
@@ -481,12 +494,25 @@ disjoin ps =
          Just False -> do
            result2 <- q
            return $
-             if expect result2 then
-               case ok result2 of
-                 Just True -> result2
-                 Just False -> result1 >>> result2
-                 Nothing -> result2
-             else expectFailureError
+             case ok result2 of
+               _ | not (expect result2) -> expectFailureError
+               Just True -> result2
+               Just False ->
+                 MkResult {
+                   ok = Just False,
+                   expect = True,
+                   reason = sep (reason result1) (reason result2),
+                   theException = theException result1 `mplus` theException result2,
+                   -- The following three fields are not important because the
+                   -- test case has failed anyway
+                   abort = False,
+                   labels = Map.empty,
+                   stamp = Set.empty,
+                   callbacks =
+                     callbacks result1 ++
+                     [PostFinalFailure Counterexample $ \st _res -> putLine (terminal st) ""] ++
+                     callbacks result2 }
+               Nothing -> result2
          Nothing -> do
            result2 <- q
            return (case ok result2 of
@@ -495,16 +521,9 @@ disjoin ps =
                      _ -> result1)
 
   expectFailureError = failed { reason = "expectFailure may not occur inside a disjunction" }
-  result1 >>> result2 | not (expect result1 && expect result2) = expectFailureError
-  result1 >>> result2 =
-    result2
-    { reason       = if null (reason result2) then reason result1 else reason result2
-    , theException = if null (reason result2) then theException result1 else theException result2
-    , stamp        = stamp result1 ++ stamp result2
-    , callbacks    = callbacks result1 ++
-                    [PostFinalFailure Counterexample $ \st _res -> putLine (terminal st) ""] ++
-                    callbacks result2
-    }
+  sep [] s = s
+  sep s [] = s
+  sep s s' = s ++ ", " ++ s'
 
 -- | Like '==', but prints a counterexample when it fails.
 infix 4 ===
